/**
 * 
 * @author Micha Pawluczuk
 *
 *	Chord is main object to handle protocol messages (asynchronous)
 *
 *	TODO
 *	- zrobic ChordConfig static ze staymi id protokou
 *	- obsuy predecessor == null
 *	- check predecessor
 *	- obsuy sytuacj gdy successor is down
 *	- klasa Entry
 */
package budrys.chord;

import java.util.Iterator;

import budrys.transport.socket.TransportClient;

public abstract class Chord implements IfcNode{
	public static final int FINGER_MAX = 63;
	public static final int FINGER_RESERVED = 64;
	public static final int PERIODICAL_DELAY = 2500;
	
	private IfcTransportServer tsPattern;
	protected IfcTransportClient tcPattern;	// clone this instance before connecting to peers
	
	private IfcTransportServer ts;	// main ts
	
	private String myIp;
	private Short myPort; 
	private String peerIp;
	private Short peerPort;
	
	protected ID myID;
	private ID predecessor;
	private ID[] fingerlist;
	private int nextFingerFix;
	
	private boolean stabilizerThreadActive;
	
	public IfcCallback onNeighbourChange;
	
	// Peers iterator
	private boolean masterLoop;
	private boolean failed;
	private Boolean waiting;
	private Peer discovery;
	
	
	public static boolean sphereIntervalBelongRightClosed(Long id, Long from, Long to){
		if (from <= to){
			return ( (from.longValue() < id.longValue()) && (id.longValue() <= to.longValue()) );
		} else {
			// from > to
			return ( (from.longValue() < id.longValue()) || (id.longValue() <= to.longValue()) );
		}
	}
	
	public static boolean sphereIntervalBelongOpen(Long id, Long from, Long to){
		if (from <= to){
			return ( (from.longValue() < id.longValue()) && (id.longValue() < to.longValue()) );
		} else {
			// from > to
			return ( (from.longValue() < id.longValue()) || (id.longValue() < to.longValue()) );
		}
	}
	
	
	/**
	 * After constructor you need to call startChord()
	 * blocking till have received his successor
	 * bootstrapping is to be done on higher level
	 * supports loopback connections
	 * 
	 * @param myIp	needed only to generate myID
	 * @param myPort
	 * @param peerIp
	 * @param peerPort
	 * @param ts
	 * @param tc
	 */
	public Chord(String myIp, Short myPort, String peerIp, Short peerPort, IfcTransportServer ts, IfcTransportClient tc){
		this.tsPattern = (IfcTransportServer) ts.cclone();
		this.tcPattern = (IfcTransportClient) tc.cclone();
		this.ts = (IfcTransportServer) this.tsPattern.cclone();
		
		this.myIp = myIp;
		this.myPort = myPort;
		this.peerIp = peerIp;
		this.peerPort = peerPort;
		
		myID = new ID(this.myIp, this.myPort);
		
		fingerlist = new ID[FINGER_RESERVED];
		fingerlist[0] = null;	// will be set later
		predecessor = null;
		nextFingerFix = 0;
		
	} // Chord
	
	/**
	 * Starting is SYNC
	 */
	public boolean startChordSync(){
		final Chord that = this;
		this.ts.startService(this.myPort, new IfcCallback(){
			
			public void success(String command, String argument, Object callback) { // on RPC receive request
				IfcCallback cb = (IfcCallback)callback;
				clog("Received request: " + (String)command + "("+(String)argument+")");
				
				// PROCESS COMMANDS
				if (command.equals("getId")){
					cb.success("getId", Chord.this.getId().toString(), null);
					
				} else if (command.equals("findSuccessor")){
					Long requestId = Long.valueOf(argument);
					Long successorFound = Chord.this.findSuccessor(requestId);
					if (successorFound != null){
						//znajduje ok ale zle zapisuje w fingerliscie
						//System.out.println("this.id:"+Chord.this.getIdHexString()+" requestId:"+(new ID(requestId))+" successorFound:"+(new ID(successorFound))+" ");
						if (successorFound != 0L) {
							clog("findSuccessor");
							cb.success("findSuccessor", successorFound.toString(), null);
						} else if (successorFound != 1L) {
							clog("findSuccessorNotFound");
							cb.success("findSuccessorNotFound", Chord.this.closestPrecedingNode(Long.valueOf(argument)).toString(), null);
						} else {
							cb.failure(new ChordException("command not understood"), null);
						}
					} else {
						// the node does not have an successor, cold boot, we are the only node in the network
						cb.success("findSuccessorNewNode", "", null);
					}
					
					
				} else if (command.equals("ping")){
					cb.success("ping", Boolean.valueOf(Chord.this.ping()).toString(), null);
					
				} else if (command.equals("closestPrecedingNode")){
					cb.success("closestPrecedingNode", Chord.this.closestPrecedingNode(Long.valueOf(argument)).toString(), null);
					
				} else if (command.equals("cNotify")){
					cb.success("cNotify", Chord.this.cNotify(Long.valueOf(argument)).toString(), null);
					
				} else if (command.equals("getPredecessor")){
					if (Chord.this.predecessor != null){
						cb.success("getPredecessor", Chord.this.getPredecessor().toString(), null);						
					} else {
						cb.success("getPredecessor", null, null);	
						//cb.failure(new ChordException("predecessor == null"), new String("PREDECESSOR_NULL"));
					}
					
				} else if (command.equals("dhtRedistribute")){
					// redistribute on demand
					if (that instanceof ChordDHT){
						((ChordDHT)Chord.this).redistributeLocalStorage(new ID(Long.valueOf(argument)));
						cb.success("dhtRedistribute", null, null);
					} else {
						clog("No such dht command: " + command + "("+argument+")");
					}
				}else if (command.equals("dhtStore")){
					// receiving any dht command means we need to operate on local storage
					if (that instanceof ChordDHT){
						ChordDHT dht = (ChordDHT) that;
						String result = dht.storeLocal((String) argument);
						cb.success("dhtStore", result, null);
					} else {
						clog("No such dht command: " + command + "("+argument+")");
					}
				} else if (command.equals("dhtGet")){
					if (that instanceof ChordDHT){
						ChordDHT dht = (ChordDHT) that;
						Log.i(Chord.this.myID+".dhtGet["+argument+"]");
						Iterator<String> it = dht.getLocal((String) argument);
						if (it!=null){
							while (it.hasNext()){
								String result = it.next();
								cb.success("dhtGet", result, null);
							}
						} else {
							cb.failure(new Exception(), (String)"NOT FOUND");
						}
						
					} else {
						clog("No such dht command: " + command + "("+argument+")");
					}
				} else if (command.equals("dhtDelete")){
					if (that instanceof ChordDHT){
						ChordDHT dht = (ChordDHT) that;
						String result = dht.deleteLocal((String) argument);
						cb.success("dhtDelete", result, null);
					} else {
						clog("No such dht command: " + command + "("+argument+")");
					}
				} else {
					clog("No such command: " + command + "("+argument+")");
				}
				
			}
			
			public void failure(Exception e, Object reason) {
				clog("Error processing RPC: " + e.toString());
			}
		});
		
		masterLoop = true;
		failed = false;
		// JOIN - find successor id
		final Peer test = new Peer(peerIp, peerPort, new IfcCallback() {
			
			public void success(String command, String argument, Object callback) {
				clog("Bootstrap peer ID: " + argument);
				if (getId().toString().equals(argument)){
					// connected to itself => creating a new p2p network
				} else {
					// connected to an existing p2p network
				}
				masterLoop = false;
			}
			
			public void failure(Exception e, Object reason) {
				//System.out.println("Wrong response");
				// TODO obsuy hacking attempt
				e.printStackTrace();
				masterLoop = false;
				failed = true;
			}
		},tcPattern);
		// SYNC
		//while(test.getId()==null){
		while(masterLoop){
			clog("Awaiting bootstrap peer.getid: "+test.getId());
			try {
				Thread.sleep(500);
			} catch (InterruptedException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
		}
		if (failed){
			return false;
		}
		// Obtain my successor
		test.findSuccessor(getId(), getId(), new IfcCallback() { //905539987707864949L
			
			public void success(String command, String argument, Object callback) {
				clog("findSuccessor: "+command+"="+argument);
				fingerlist[0] = new ID(Long.valueOf(argument));
			}
			
			public void failure(Exception e, Object reason) {
				clog("Connected to itself - creating a new chord network");
				fingerlist[0] = new ID(myID.getID());
				
			}
		});
		// SYNC
		while(fingerlist[0]==null){
			try {
				clog("Awaiting my.successorid: "+fingerlist[0]);
				Thread.sleep(500);
			} catch (InterruptedException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
		}
		clog("Joined the network, successor = " + fingerlist[0]);
		
		// NOTIFY
		Peer successorx = new Peer(fingerlist[0].getID(), tcPattern); 
		successorx.cNotify(Chord.this.myID.getID(), new IfcCallback() {
			
			public void success(String command, String argument, Object callback) {
				// no callback
			}
			
			public void failure(Exception e, Object reason) {
				// no callback
			}
		});
		// PERIODICALLY CALLED FUNCTIONS
		(new Thread(){
			public void run() {
				Chord.this.stabilizerThreadActive = true;
				while (Chord.this.stabilizerThreadActive){
					try {
						Thread.sleep(PERIODICAL_DELAY);
					} catch (InterruptedException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
					Chord.this.stabilize();
					Chord.this.fixFingers();
					Chord.this.checkPredecessor();
				}
			}
		}).start();
		return true;
	} // startChordSync() SYNC
	
	/**
	 * Starting is ASYNC
	 */
	public void startChordAsync(final IfcCallback callback){
		(new Thread(){
			public void run(){
				boolean status = startChordSync();
				if (callback != null){
					if (status == true){
						callback.success("OK", "", null);
					} else {
						callback.failure(new ChordException("FAILED"), "");
					}
				}
			}
		}).start();
	}
	
	/**
	 * LOCAL
	 */
	public void stopChord(){
		if (this instanceof ChordDHT){
			((ChordDHT)this).onLeaveDHT();
		}
		// no need to wait for asynchronous request
		this.ts.stopService();
		this.stabilizerThreadActive = false;
	}
	
	public boolean isRuning(){
		return this.stabilizerThreadActive;
	}
	
	public String getMyIp() {
		return myIp;
	}
	
	// CHORD IMPLEMENTATION
	

	
	/**
	 * LOCAL
	 */
	public Long getId() {
		return myID.getID();
	}
	
	/**
	 * LOCAL
	 */
	public String getIdHexString() {
		return myID.toString();
	}
	
	/**
	 * LOCAL
	 * @return
	 */
	public Long getPredecessor() {
		if (predecessor == null){
			clog("My predecessor: null");
			return null;
		} else {
			clog("My predecessor: "+predecessor.getID());		
			return predecessor.getID();
		}
	}
	
	/**
	 * LOCAL
	 * @return
	 */
	public ID getPredecessorID() {	
		return predecessor;
	}
	
	/**
	 * LOCAL
	 * @return
	 */
	public ID getSuccessor() {
		return fingerlist[0];
	}
	
	/**
	 * LOCAL
	 * TODO change to some proof based method
	 * @return
	 */
	private boolean ping() {
		return true;	// the fact the method has returned is a proof of being alive
	}

	/**
	 * LOCAL
	 * @param id
	 * @return 0 = forward query around the circle, null = no successors noone in fingerlist, other value = successor of id
	 */
	private Long findSuccessor(Long id) {
		if (fingerlist[0] == null){
			// we are the only node in the network
			return null;	// impossible to obtain
		}
		ID successor = fingerlist[0];
		if (sphereIntervalBelongRightClosed(id, getId(), successor.getID())){
			// found
			return successor.getID();
		} else {
			//	 Forward query around the circle
			return 0L;	// impossible to have an ID == 0 thus return not in my range, A NIE 1?
			//return closestPrecedingNode(id);	// wrong response, but to be corrected iteratively
		}
	}

	/**
	 * LOCAL
	 * @param id
	 * @return
	 */
	private Long closestPrecedingNode(Long id) {
		ID successor = fingerlist[0];
		if ((fingerlist.length <= 0)){
			// TODO tutaj successor == null
			return successor.getID();
		} else {
			for(int i=fingerlist.length-1; i>=0; --i){
				//clog("iter: "+i);
				if (fingerlist[i] != null){
					//if ((getId() < fingerlist[i].getID()) && (fingerlist[i].getID() < id)){
					if (sphereIntervalBelongOpen(fingerlist[i].getID(), getId(), id)){
						return fingerlist[i].getID();
					}
				}
			}
		}
		//return getId();
		return successor.getID();
	}

	/**
	 * LOCAL
	 * Notify
	 * @param id
	 * @return
	 */
	private Boolean cNotify(Long id) {
		
		clog("cNotify predecessor: "+(this.predecessor==null ? "null" : this.predecessor.getID())+" id: "+id);
		if (  
				( this.predecessor == null )   
			||  
				//( (predecessor.getID().longValue() < id.longValue()) && (id.longValue() < this.myID.getID().longValue()) )
				( sphereIntervalBelongOpen(id, this.predecessor.getID(), this.myID.getID()) )
			||
				( this.myID.getID().equals(this.predecessor.getID()) && !this.myID.getID().equals(id))
			){
			this.predecessor = new ID(id);
			if (this instanceof ChordDHT){
				((ChordDHT)Chord.this).redistributeLocalStorage(this.predecessor);
			}
			Chord.this.callOnNeighbourChange();
			clog("return true;");
			return true;
		}
		return false;
	}
	
	/**
	 * ASYNC
	 * to be called periodically
	 */
	private void stabilize(){
		final ID successor = fingerlist[0];
		final Peer successorx = new Peer(successor.getID(), tcPattern);
		successorx.getPredecessor(new IfcCallback() {
			
			public void success(String command, final String argument, Object callback) {
				if (command.equals("getPredecessor")){
					Peer newSuccessorx = successorx;
					if (argument != null){
						Long succPred = Long.valueOf(argument);
						//if (  (Chord.this.getId().longValue() < succPred.longValue())  &&  (succPred.longValue() < successor.getID().longValue())  ){
						if (
								( sphereIntervalBelongOpen(succPred, Chord.this.getId(), successor.getID()) )
							||
								( successor.getID().equals(Chord.this.getId()) )
							){
							Chord.this.fingerlist[0] = new ID(succPred);
							Chord.this.callOnNeighbourChange();
							clog("Changed my successor to: "+succPred);
							newSuccessorx = new Peer(succPred, tcPattern);
						}
					} else {	// argument == null
						
					}
					// this is not a critical operation for us
					newSuccessorx.cNotify(Chord.this.myID.getID(), new IfcCallback() {
						
						public void success(String command, String cargument, Object callback) {
							// TODO Auto-generated method stub
							clog("I have notified successful: "+successorx.getId() +" boolean: "+cargument);
//							if (cargument=="true"){
//								if (Chord.this instanceof ChordDHT){
//									((ChordDHT)Chord.this).redistributeLocalStorage();
//								}
//							}
						}
						
						public void failure(Exception e, Object reason) {
							//Log.i("notify failure:");
							//e.printStackTrace();
						}
					});
				} else {
					// cannot getPredecessor
					clog("Cannot getPredecessor:");
				}
			}
			
			public void failure(Exception e, Object reason) {
				// our successor has failed, or does not have
				Log.i("getPredecessor failure: stabilize:"+Log.idFormat(Chord.this.getId())+" peer:"+Log.idFormat(successorx.getId()));
				removeFailedID(successorx.getId());
				System.out.println( debugInfoJSON() );
				//e.printStackTrace();
			}
		});
	}
	
	/**
	 * ASYNC
	 * to be called periodically
	 */
	private void fixFingers(){
		this.nextFingerFix += 1;
		if (this.nextFingerFix > FINGER_MAX){
			this.nextFingerFix = 1;
		}
		
		final Long idToFind = (long) (getId() + Math.pow(2, this.nextFingerFix -1));
		Peer requestPeer = new Peer(this.getId(), tcPattern);	// lepsze

		// recursive
		requestPeer.findSuccessor(idToFind, getId(), new IfcCallback() {
			
			public void success(String command, String argument, Object callback) {
				Chord.this.fingerlist[Chord.this.nextFingerFix] = new ID(Long.valueOf(argument));
			}
			
			public void failure(Exception e, Object reason) {
				// peer has failed
				removeFailedID((Long)reason);
				
				Log.i("fixFingers findSuccessor failure on peer: "+Log.idFormat((Long)reason)+"; finger:"+Chord.this.nextFingerFix+"; node:"+Log.idFormat(Chord.this.getId())+"; idToFind:"+Log.idFormat(idToFind));
				//debugInfoJSON();
				//e.printStackTrace();
			}
		});
	}
	
	/**
	 * LOCAL
	 * covers case when successor failed 
	 */
	private void removeFailedID(long failedId){
		System.out.println("removed: "+(new ID(failedId)).toString());
		for (int i=0; i<FINGER_RESERVED; ++i){
			if (fingerlist[i]!=null){
				if (fingerlist[i].getID().equals(failedId)){
					fingerlist[i] = null;
				}
			}
		}
		if (fingerlist[0] == null){		// if successor has failed
			for (int i=1; i<FINGER_RESERVED; ++i){
				if (fingerlist[i]!=null){
					fingerlist[0] = fingerlist[i];
				}
			}
			Chord.this.callOnNeighbourChange();
		}
		if (fingerlist[0] == null){		// there are no more fingers on fingerlist 
			fingerlist[0] = this.myID;
			Chord.this.callOnNeighbourChange();
		}
	}
	
	/**
	 * ASYNC
	 * Check whether predecessor has failed
	 * to be called periodically
	 */
	private void checkPredecessor(){
		if (predecessor == null){
			return ;
		}
		Peer pred = new Peer(predecessor.getID(), tcPattern);
		pred.ping(new IfcCallback() {
			
			public void success(String command, String argument, Object callback) {
				// is alive
			}
			
			public void failure(Exception e, Object reason) {
				predecessor = null;
				Chord.this.callOnNeighbourChange();
			}
		});
	}
	
	//private void detectAnomalia(){
		// BUDRYS
		/* check if 
		 * 		successor = myid
		 * 		all fingers = myid
		 * 		predecessor = null
		 * then
		 * 		reconnect because of mass failure
		 * else
		 * 		if successor fail
		 * 		then
		 * 			successor = next valid finger
		 * 		end
		 * end
		*/
	//}
	
	
	// events
	private void callOnNeighbourChange(){
		if (this.onNeighbourChange != null){
			(new Thread(){
				public void run() {
					String s = (Chord.this.getSuccessor() == null) ? null : Chord.this.getSuccessor().toString();
					String p = (Chord.this.getPredecessor() == null) ? null : Chord.this.getPredecessorID().toString();
					Chord.this.onNeighbourChange.success(s, p, null);
				}
			}).start();
		}
	}
	
	// diagnostics
	
	/**
	 * ASYNC
	 * Loop through all peers
	 */
	public void loopThroughPeers(final IfcCallback each, final IfcCallback finish){
		if (this.isRuning()){
			/*if (this.getSuccessor() == null){
				finish.success("Chord no successor", "", null);
				return;
			}
			loopThroughPeersLocalSync(this.getSuccessor(), each, finish);*/
			(new Thread(){
				public void run() {
					loopThroughPeersLocalSync(Chord.this.myID, each, finish);
				}
			}).start();
		} else {
			// not runing
			finish.success("Chord not runing", "", null);
		}
	}
	private void loopThroughPeersLocalSync(final ID start, final IfcCallback each, final IfcCallback finish){
		this.masterLoop = true;
		this.discovery = new Peer(start.getID(), new TransportClient());
		while(masterLoop){
			this.waiting = true;
			discovery.findSuccessor(1L + discovery.getPeerIdOffline().getID(), discovery.getPeerIdOffline().getID(), new IfcCallback() {
				public void success(String command, String argument, Object callback) {
					final ID next = (new ID(Long.valueOf(argument)));
					Chord.this.waiting = false;
					Chord.this.discovery = new Peer(next.getID(), new TransportClient());
					(new Thread(){
						public void run() {
							each.success("NEXT", next.getID().toString(), null);
						}
					}).start();
				}
				public void failure(Exception e, Object reason) {
					Chord.this.waiting = false;
					Chord.this.masterLoop = false;
					finish.failure(e, "Error: "+discovery.getPeerIdOffline().toString());
				}
			});
			while (waiting) {
				try {
					Thread.sleep(300);
				} catch (InterruptedException e1) {
					e1.printStackTrace();
				}
			}
			if (start.getID().equals(discovery.getPeerIdOffline().getID())){
				Chord.this.masterLoop = false;
				finish.success("FINISH", "", null);
			}
		}	// masterLoop
	}
	/**
	 * Async version is very unstable
	 * @param start
	 * @param current
	 * @param each
	 * @param finish
	 */
	/*private void loopThroughPeersLocalAsync(final ID start, ID current, final IfcCallback each, final IfcCallback finish){
		final Peer discovery = new Peer(current.getID(), new TransportClient());
		//each.success("PRE NEXT", current.toString(), null);
		discovery.findSuccessor(1L+discovery.getPeerIdOffline().getID(), discovery.getPeerIdOffline().getID(), new IfcCallback() {
			public void success(String command, String argument, Object callback) {
				Log.i("success");
				ID nextPeer = new ID(Long.valueOf(argument));
				each.success("NEXT", nextPeer.toString(), null);
				if (nextPeer.getID().equals(start.getID())){
					// finish
					finish.success("FINISH", "", null);
					return;
				} else {
					loopThroughPeersLocalAsync(start, nextPeer, each, finish);
				}
			}
			public void failure(Exception e, Object reason) {
				finish.failure(e, "Error: "+discovery.getPeerIdOffline().toString());
			}
		});
	}*/
	
	
	// DEBUG
	/**
	 * Comment System.out to stop debug output
	 * @param s
	 */
	private synchronized void clog(String s){
		//System.out.println(getId() + " " + s);
	}
	
	public synchronized String debugInfo(){
		return "ID: "+myID+"; successor: "+fingerlist[0]+"; predecessor: "+predecessor;
	}
	public synchronized String debugInfoJSON(){
		
		StringBuilder str = new StringBuilder();
		boolean first = true;
		for (int i=0; i<64; ++i){
			if (fingerlist[i] != null){
				if (!first){
					str.append(", ");
				} else {
					first = false;
				}
				//final Long idToFind = (long) (getId() + Math.pow(2, this.nextFingerFix -1));
				Long succOfKey = (long) (getId() + (long)Math.pow(2, i -1));
				ID succOfKeyID = new ID(succOfKey);
				//str.append("{\""+i+"\":\""+fingerlist[i]+"\"}");
				str.append("{\""+succOfKeyID+"\":\""+fingerlist[i]+"\"}");
			}
		}
		return "{\"id\":\""+myID+"\", \"succ\":\""+fingerlist[0]+"\", \"pred\":\""+predecessor+"\", \"fingerFix\":\""+String.format("%02d", this.nextFingerFix)+"\", \"fingers\":["+str.toString()+"]}";
	}
//	public synchronized String debugInfoGraph(){
//		
//		StringBuilder str = new StringBuilder();
//		str.append("digraph G {\n"+
//	"node [label=\"\\N\"];\n"+
//	"graph [bb=\"0,0,406,394\",\n"+
//	"	_draw_=\"c 5 -white C 5 -white P 4 0 0 0 394 406 394 406 0 \",\n"+
//	"	xdotversion=\"1.2\"];\n"+
//	"subgraph cluster_c0 {\n"+
//	"	graph [bb=\"\",\n"+
//	"		_draw_=\"\",\n"+
//	"		xdotversion=\"\"];\n");
//		boolean first = true;
//		for (int i=0; i<64; ++i){
//			if (fingerlist[i] != null){
//				if (!first){
//					str.append(", ");
//				} else {
//					first = false;
//				}
//				//final Long idToFind = (long) (getId() + Math.pow(2, this.nextFingerFix -1));
//				Long succOfKey = (long) (getId() + (long)Math.pow(2, i -1));
//				ID succOfKeyID = new ID(succOfKey);
//				//str.append("{\""+i+"\":\""+fingerlist[i]+"\"}");
//				str.append("{\""+succOfKeyID+"\":\""+fingerlist[i]+"\"}");
//			}
//		}
//		str.append("}");
//		return "{\"id\":\""+myID+"\", \"succ\":\""+fingerlist[0]+"\", \"pred\":\""+predecessor+"\", \"fingerFix\":\""+String.format("%02d", this.nextFingerFix)+"\", \"fingers\":["+str.toString()+"]}";
//	}

}
